# -*- coding: utf-8 -*-
"""
    Media Server module implementing the MediaRequest API
    
    Adds a Flash socket policy file server for
    WebSocket requests emulated via Flash
"""

from __future__ import with_statement
from base64 import b64decode
from contextlib import closing
import simplejson as json
import logging
import os
import re
from StringIO import StringIO
from twisted.internet.protocol import (
    Protocol,
    Factory
)
from twisted.web import resource
from twisted.web.static import File
from twisted.internet import task
from threading import (
    Thread,
    Timer
)
import urllib2

from websocket import WebSocketHandler, WebSocketSite

logging.basicConfig(level=logging.INFO, format='%(levelname)-8s: %(message)s', datefmt='%H:%M:%S')
logger = logging.getLogger("suika")
logger.setLevel(logging.INFO)

# General constants
MEDIA_URL = 'http://dtm/_brodala/homepage/suika/media/'
MEDIA_PATH = os.path.abspath('media/')
MEDIA_REQUEST_VERSION = 1.0
MEDIA_REQUEST_INFO = 'info'
MEDIA_REQUEST_LIST = 'list'
MEDIA_REQUEST_ADD = 'add'
MEDIA_REQUEST_REMOVE = 'remove'

UPLOAD_USER = 'suika'
# SHA-1 password hash
UPLOAD_PASSWORD = '712e9b0a3b61937ec72352bf49d9c04090c9d28b' # suika

# Error constants
ERROR_UNKNOWN = 'unknown'
ERROR_UNKNOWN_PROTOCOL_VERSION = 'unknownProtocolVersion'
ERROR_UNKNOWN_REQUEST_TYPE = 'unknownRequestType'
ERROR_AUTHENTICATION_REQUIRED = 'authenticationRequired'
ERROR_AUTHENTICATION_FAILED = 'authenticationFailed'
ERROR_MALFORMED_REQUEST = 'malformedRequest'

# Other constants
UID_WWW_DATA = 33
UID_NOBODY = 65534

__author__ = 'Mathias Brodala'

class MediaRequestHandler(WebSocketHandler):
    """ 
        WebSocket handler for MediaRequest requests
    """
    def __init__(self, transport):
        WebSocketHandler.__init__(self, transport)

    def __del__(self):
        logger.info("Deleting WebSocket handler")
        
    def send_resource_added_message(self):
        """
            Sends a message notifying about an added resource
        """
        response = {
            'version': MEDIA_REQUEST_VERSION,
            'type': MEDIA_REQUEST_ADD,
            'result': 'success',
            'resources': {
                MEDIA_URL + 'image/fractal-rainbow-ocean.jpg': {
                    'title': 'Fraktaler Regenbogen-Ozean',
                    'type': 'image/jpeg',
                    'size': 138301,
                    'thumbnail': MEDIA_URL + 'image/fractal-rainbow-ocean-thumb.jpg'
                }
            }
        }
        
        logger.info('Response: %s' % response)
        
        self.transport.write(json.dumps(response))

    def frameReceived(self, frame):
        """
            Handles MediaRequest requests
        """
        response = {
            'version': MEDIA_REQUEST_VERSION
        }
        logger.info('Message from %s: %s' % (self.transport.getPeer(), frame))
        
        try:
            request = json.loads(frame)['mediaRequest']
        except (ValueError, KeyError), e:
            response.update({
                'result': 'error',
                'error': {
                    'code': ERROR_MALFORMED_REQUEST,
                    'message': 'Processing of the request failed'
                }
            })
            self.transport.write(json.dumps(response))
            return
        
        logger.info('Request: %s' % request)

        response['type'] = request['type']
        
        if 'id' in request:
            response['requestId'] = request['id']
            
        if 'version' not in request or request['version'] != MEDIA_REQUEST_VERSION:
            response.update({
                'result': 'error',
                'error': {
                    'code': ERROR_UNKNOWN_PROTOCOL_VERSION,
                    'message': 'Unknown protocol version'
                }
            })
            self.transport.write(json.dumps(response))
            return
            
        response.update(getattr(
            self,
            "handle_%s_request" % request['type'],
            self.handle_unknown_request
        )(request))
        
        logger.info('Response: %s' % response)
        
        self.transport.write(json.dumps(response))

    def connectionMade(self):
        logger.info("Connected to client")

    def connectionLost(self, reason):
        logger.info("Disconnected from client")
        
    def handle_info_request(self, request):
        """
            Handles requests of the type MediaRequest.info
        """
        return {
            'result': 'success',
            'info': {
                'title': 'Suika',
                'allowed': ['list', 'add'],
                'restricted': ['add']
            }
        }
        
    def handle_list_request(self, request):
        """
            Handles requests of the type MediaRequest.list
        """
        timer = Timer(5, self.send_resource_added_message)
        timer.daemon = True
        timer.start()
        
        return {
            'result': 'success',
            'resources': {
                MEDIA_URL + 'image/watermelon.jpg': {
                    'title': 'Wassermelone',
                    'type': 'image/jpeg',
                    'size': 33793,
                    'thumbnail': MEDIA_URL + 'image/watermelon-thumb.jpg'
                },
                MEDIA_URL + 'image/waterfall.jpg': {
                    'title': 'Wasserfall',
                    'type': 'image/jpeg',
                    'size': 222272,
                    'thumbnail': MEDIA_URL + 'image/waterfall-thumb.jpg'
                },
                MEDIA_URL + 'image/dawn.jpg': {
                    'title': 'Dämmerung',
                    'type': 'image/jpeg',
                    'size': 435566,
                    'thumbnail': MEDIA_URL + 'image/dawn-thumb.jpg'
                }
            }
        }
        
    def handle_add_request(self, request):
        """
            Handles requests of the type MediaRequest.add
        """
        if 'credentials' not in request or \
           'username' not in request['credentials'] or \
           'password' not in request['credentials']:
            return {
                'result': 'error',
                'error': {
                    'code': ERROR_AUTHENTICATION_REQUIRED,
                    'message': 'Authentication required.'
                }
            }
        
        if request['credentials']['username'] != UPLOAD_USER or \
           request['credentials']['password'] != UPLOAD_PASSWORD:
            return {
                'result': 'error',
                'error': {
                    'code': ERROR_AUTHENTICATION_FAILED,
                    'message': 'Authentication failed.'
                }
            }

        request_resources = []
        resources = {}
          
        if 'resources' in request:
            request_resources += request['resources']
            
            for url, properties in request['resources'].iteritems():
                try:
                    general_type, specific_type = properties['type'].split('/')
                except:
                    general_type = 'application'
                    
                directory_path = os.path.join(MEDIA_PATH, general_type)
                
                if not os.path.exists(directory_path):
                    try:
                        logger.info('Trying to create directory for resources: %s' % directory_path)
                        os.mkdir(directory_path, 755)
                    except:
                        logger.warning('Failed to create directory for resources')
                        break
                
                file_name = url.split('/')[-1]
                file_path = os.path.join(MEDIA_PATH, general_type, file_name)
                
                try:
                    with closing(urllib2.urlopen(url)) as infile:
                        with closing(open(file_path, 'w+')) as outfile:
                            outfile.write(infile.read())
                except:
                    logger.warning('Failed to read resource data')
                    break
                
                url = '%s%s/%s' % (MEDIA_URL, general_type, file_name)
                resources[url] = properties
          
        if 'files' in request:
            request_resources += request['files']
            
            for file in request['files']:
                try:
                    general_type, specific_type = file['type'].split('/')
                except:
                    general_type = 'application'
                    
                directory_path = os.path.join(MEDIA_PATH, general_type)
                
                if not os.path.exists(directory_path):
                    try:
                        logger.info('Trying to create directory for files: %s' % directory_path)
                        os.mkdir(directory_path, 755)
                    except:
                        logger.warning('Failed to create directory for files')
                        break
                
                file_path = os.path.join(MEDIA_PATH, general_type, file['name'])
                
                try:
                    with closing(StringIO(b64decode(file['data']))) as infile:
                        with closing(open(file_path, 'w+')) as outfile:
                            outfile.write(infile.read())
                except:
                    logger.warning('Failed to read file data')
                    break
                
                url = '%s%s/%s' % (MEDIA_URL, general_type, file['name'])
                resources[url] = {
                    'size': file['size'],
                    'type': file['type']
                }
        
        if len(resources) == len(request_resources):
            return {
                'result': 'success',
                'resources': resources
            }

        return {
            'result': 'error',
            'error': {
                'code': ERROR_UNKNOWN,
                'message': 'Failed to add some resources.'
            }
        }
        
    def handle_unknown_request(self, request):
        """
            Handles unknown MediaRequest requests
        """
        return {
            'result': 'error',
            'error': {
                'code': ERROR_UNKNOWN_REQUEST_TYPE,
                'message': 'Unknown request type'
            }
        }

class FlashSocketPolicy(Protocol):
    """
        WebSocket handler for Flash socket policy file requests
        
        See: https://github.com/gimite/web-socket-js
        See: http://www.adobe.com/devnet/flashplayer/articles/socket_policy_files.html
    """
    def connectionMade(self):
        policy = '''<?xml version="1.0"?>
<!DOCTYPE cross-domain-policy SYSTEM "http://www.macromedia.com/xml/dtds/cross-domain-policy.dtd">
<cross-domain-policy>
  <allow-access-from domain="*" to-ports="*" />
</cross-domain-policy>'''
        self.transport.write(policy)
        self.transport.loseConnection()

if __name__ == "__main__":
    from twisted.internet import reactor
    from twisted.internet.error import CannotListenError
    
    # Run MediaRequest requests handler
    root = File('.')
    site = WebSocketSite(root)
    site.addHandler('/suika', MediaRequestHandler)
    reactor.listenTCP(1111, site)
    
    # Run Flash socket policy file requests handler
    factory = Factory()
    factory.protocol = FlashSocketPolicy
    
    try:
        reactor.listenTCP(843, factory)
    except CannotListenError:
        logger.warning('Failed to start the Flash socket policy file '
                       'server on port 843; are you running as root?')
    else:
        # Move to unprivileged user
        import os
        os.seteuid(UID_WWW_DATA)

    reactor.run()
